@@ -31,8 +31,16 @@ static EXTENSION_FIXED_NAMES: [&str; 6] = [
     // "extraRootProperties",
 ];
 
-pub static CITYJSON_VERSIONS: [&str; 2] = ["1.0.3", "1.1.0"];
+static CITYJSON_V10_VERSION: &str = "1.0.3";
 
+#[derive(Serialize, Deserialize, Debug)]
+struct GeomMPo {
+    boundaries: Vec<usize>,
+}
+#[derive(Serialize, Deserialize, Debug)]
+struct GeomMLS {
+    boundaries: Vec<Vec<usize>>,
+}
 #[derive(Serialize, Deserialize, Debug)]
 struct GeomMSu {
     boundaries: Vec<Vec<Vec<usize>>>,
@@ -52,12 +60,26 @@ struct Doc {
     CityObjects: HashMap<String, Value>,
 }
 
+pub fn get_cityjson_schema_all_versions() -> Vec<String> {
+    let mut l: Vec<String> = Vec::new();
+    //-- v1.0
+    l.push(CITYJSON_V10_VERSION.to_string());
+    //-- v1.1
+    let schema_str = include_str!("../schemas/11/cityjson.min.schema.json");
+    let schema: Value = serde_json::from_str(schema_str).unwrap();
+    let vs = &schema["$id"].to_string();
+    l.push(vs.get(34..39).unwrap().to_string());
+    l
+}
+
 #[derive(Debug)]
 pub struct CJValidator {
     j: Value,
+    jschema: Value,
     jexts: HashMap<String, Value>,
     duplicate_keys: bool,
-    version: i32,
+    version_file: i32,
+    version_schema: String,
 }
 
 impl CJValidator {
@@ -65,9 +87,11 @@ impl CJValidator {
         let l: HashMap<String, Value> = HashMap::new();
         let mut v = CJValidator {
             j: json!(null),
+            jschema: json!(null),
             jexts: l,
             duplicate_keys: false,
-            version: 0,
+            version_file: 0,
+            version_schema: "none".to_string(),
         };
         //-- parse the dataset and convert to JSON
         let re: Result<Value, _> = serde_json::from_str(&str_dataset);
@@ -79,9 +103,17 @@ impl CJValidator {
         v.j = j;
         //-- check cityjson version
         if v.j["version"] == "1.1" {
-            v.version = 11;
+            v.version_file = 11;
+            let schema_str = include_str!("../schemas/11/cityjson.min.schema.json");
+            v.jschema = serde_json::from_str(schema_str).unwrap();
+            let vs = &v.jschema["$id"].to_string();
+            v.version_schema = vs.get(34..39).unwrap().to_string();
         } else if v.j["version"] == "1.0" {
-            v.version = 10;
+            v.version_file = 10;
+            let schema_str = include_str!("../schemas/11/cityjson.min.schema.json");
+            v.jschema = serde_json::from_str(schema_str).unwrap();
+            let vs = &v.jschema["$id"].to_string();
+            v.version_schema = vs.get(34..39).unwrap().to_string();
         }
         //-- check for duplicate keys in CO object
         let re: Result<Doc, _> = serde_json::from_str(&str_dataset);
@@ -127,7 +159,11 @@ impl CJValidator {
     }
 
     pub fn get_input_cityjson_version(&self) -> i32 {
-        self.version
+        self.version_file
+    }
+
+    pub fn get_cityjson_schema_version(&self) -> String {
+        self.version_schema.to_owned()
     }
 
     pub fn validate_schema(&self) -> Vec<String> {
@@ -135,22 +171,17 @@ impl CJValidator {
             return vec!["Not a valid JSON file".to_string()];
         }
         //-- which cityjson version
-        if self.version == 0 {
+        if self.version_file == 0 {
             let s: String = format!(
                 "CityJSON version {} not supported [only \"1.0\" and \"1.1\"]",
                 self.j["version"]
             );
             return vec![s];
         }
-        //-- fetch the correct schema
-        let mut schema_str = include_str!("../schemas/10/cityjson.min.schema.json");
-        if self.version == 11 {
-            schema_str = include_str!("../schemas/11/cityjson.min.schema.json");
-        }
-        let schema = serde_json::from_str(schema_str).unwrap();
+
         let compiled = JSONSchema::options()
             .with_draft(Draft::Draft7)
-            .compile(&schema)
+            .compile(&self.jschema)
             .expect("A valid schema");
         let result = compiled.validate(&self.j);
         let mut ls_errors: Vec<String> = Vec::new();
@@ -531,27 +562,75 @@ impl CJValidator {
         let mut ls_errors: Vec<String> = Vec::new();
         let cos = self.j.get("CityObjects").unwrap().as_object().unwrap();
         for key in cos.keys() {
+            //-- check geometry
             let x = self.j["CityObjects"][key]["geometry"].as_array();
             if x.is_some() {
                 for g in x.unwrap() {
-                    if g["type"] == "MultiSurface" || g["type"] == "CompositeSurface" {
-                        let aa: GeomMSu = serde_json::from_value(g.clone()).unwrap();
-                        let re = above_max_index_msu(&aa.boundaries, max_index);
+                    if g["type"] == "MultiPoint" {
+                        let a: GeomMPo = serde_json::from_value(g.clone()).unwrap();
+                        for each in a.boundaries {
+                            if each >= max_index {
+                                let s2 = format!("Vertices {} don't exist", each);
+                                ls_errors.push(s2);
+                            }
+                        }
+                    } else if g["type"] == "MultiLineString" {
+                        let a: GeomMLS = serde_json::from_value(g.clone()).unwrap();
+                        for l in a.boundaries {
+                            for each in l {
+                                if each >= max_index {
+                                    let s2 = format!("Vertices {} don't exist", each);
+                                    ls_errors.push(s2);
+                                }
+                            }
+                        }
+                    } else if g["type"] == "MultiSurface" || g["type"] == "CompositeSurface" {
+                        let a: GeomMSu = serde_json::from_value(g.clone()).unwrap();
+                        let re = above_max_index_msu(&a.boundaries, max_index);
                         if re.is_err() {
                             ls_errors.push(re.err().unwrap());
                         }
                     } else if g["type"] == "Solid" {
-                        let aa: GeomSol = serde_json::from_value(g.clone()).unwrap();
-                        let re = above_max_index_sol(&aa.boundaries, max_index);
+                        let a: GeomSol = serde_json::from_value(g.clone()).unwrap();
+                        let re = above_max_index_sol(&a.boundaries, max_index);
                         if re.is_err() {
                             ls_errors.push(re.err().unwrap());
                         }
                     } else if g["type"] == "MultiSolid" || g["type"] == "CompositeSolid" {
-                        let aa: GeomMSol = serde_json::from_value(g.clone()).unwrap();
-                        let re = above_max_index_msol(&aa.boundaries, max_index);
+                        let a: GeomMSol = serde_json::from_value(g.clone()).unwrap();
+                        let re = above_max_index_msol(&a.boundaries, max_index);
                         if re.is_err() {
                             ls_errors.push(re.err().unwrap());
                         }
+                    } else if g["type"] == "GeometryInstance" {
+                        let a: GeomMPo = serde_json::from_value(g.clone()).unwrap();
+                        for each in a.boundaries {
+                            if each >= max_index {
+                                let s2 = format!("Vertex {} doesn't exist (in #{})", each, key);
+                                ls_errors.push(s2);
+                            }
+                        }
+                    }
+                }
+            }
+            //-- check address
+            if self.j["CityObjects"][key]["type"] == "Building"
+                || self.j["CityObjects"][key]["type"] == "BuildingPart"
+                || self.j["CityObjects"][key]["type"] == "BuildingUnit"
+                || self.j["CityObjects"][key]["type"] == "Bridge"
+                || self.j["CityObjects"][key]["type"] == "BridgePart"
+            {
+                let x = self.j["CityObjects"][key]["address"].as_array();
+                if x.is_some() {
+                    for ad in x.unwrap() {
+                        let t = ad.pointer("/location/boundaries");
+                        if t.is_some() {
+                            let i = t.unwrap().get(0).unwrap().as_u64().unwrap();
+                            if (i as usize) >= max_index {
+                                let s2 = format!("Vertices {} don't exist", i);
+                                ls_errors.push(s2);
+                            }
+                        }
                     }
                 }
             }
@@ -563,12 +642,25 @@ impl CJValidator {
         let mut ls_errors: Vec<String> = Vec::new();
         let mut uniques: HashSet<usize> = HashSet::new();
         let cos = self.j.get("CityObjects").unwrap().as_object().unwrap();
-        for theid in cos.keys() {
-            let x = self.j["CityObjects"][theid]["geometry"].as_array();
+        for key in cos.keys() {
+            //-- check geometry
+            let x = self.j["CityObjects"][key]["geometry"].as_array();
             if x.is_some() {
                 let gs = x.unwrap();
                 for g in gs {
-                    if g["type"] == "MultiSurface" || g["type"] == "CompositeSurface" {
+                    if g["type"] == "MultiPoint" {
+                        let a: GeomMPo = serde_json::from_value(g.clone()).unwrap();
+                        for each in a.boundaries {
+                            uniques.insert(each);
+                        }
+                    } else if g["type"] == "MultiLineString" {
+                        let a: GeomMLS = serde_json::from_value(g.clone()).unwrap();
+                        for l in a.boundaries {
+                            for each in l {
+                                uniques.insert(each);
+                            }
+                        }
+                    } else if g["type"] == "MultiSurface" || g["type"] == "CompositeSurface" {
                         let gv: GeomMSu = serde_json::from_value(g.clone()).unwrap();
                         collect_indices_msu(&gv.boundaries, &mut uniques);
                     } else if g["type"] == "Solid" {
@@ -577,6 +669,29 @@ impl CJValidator {
                     } else if g["type"] == "MultiSolid" || g["type"] == "CompositeSolid" {
                         let gv: GeomMSol = serde_json::from_value(g.clone()).unwrap();
                         collect_indices_msol(&gv.boundaries, &mut uniques);
+                    } else if g["type"] == "GeometryInstance" {
+                        let a: GeomMPo = serde_json::from_value(g.clone()).unwrap();
+                        for each in a.boundaries {
+                            uniques.insert(each);
+                        }
+                    }
+                }
+            }
+            //-- check address
+            if self.j["CityObjects"][key]["type"] == "Building"
+                || self.j["CityObjects"][key]["type"] == "BuildingPart"
+                || self.j["CityObjects"][key]["type"] == "BuildingUnit"
+                || self.j["CityObjects"][key]["type"] == "Bridge"
+                || self.j["CityObjects"][key]["type"] == "BridgePart"
+            {
+                let x = self.j["CityObjects"][key]["address"].as_array();
+                if x.is_some() {
+                    for ad in x.unwrap() {
+                        let t = ad.pointer("/location/boundaries");
+                        if t.is_some() {
+                            let i = t.unwrap().get(0).unwrap().as_u64().unwrap();
+                            uniques.insert(i as usize);
+                        }
                     }
                 }
             }
@@ -589,7 +704,7 @@ impl CJValidator {
                 let total = self.j["vertices"].as_array().unwrap().len();
                 for each in 0..total {
                     if !uniques.contains(&each) {
-                        ls_errors.push(format!("vertex #{} is unused", each));
+                        ls_errors.push(format!("Vertex #{} is unused", each));
                     }
                 }
             }
@@ -609,7 +724,11 @@ impl CJValidator {
                     if g.get("semantics").is_none() {
                         continue;
                     }
-                    if g["type"] == "MultiSurface" || g["type"] == "CompositeSurface" {
+                    if g["type"] == "MultiPoint"
+                        || g["type"] == "MultiLineString"
+                        || g["type"] == "MultiSurface"
+                        || g["type"] == "CompositeSurface"
+                    {
                         //-- length of the sem-surfaces == # of surfaces
                         if g["boundaries"].as_array().unwrap().len()
                             != g["semantics"]["values"].as_array().unwrap().len()
